""" Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import ast
import json
import requests
import logging
from django.conf import settings
from cshmac.requests import HmacAuth
from .requests_auth import get_requests_auth
from urllib.parse import urlparse

logger = logging.getLogger('connectors')


def _get_request_params(method, body=None, *args, **kwargs):
    env = kwargs.get('env', {})
    bodyless_methods = ['head', 'get']

    if not isinstance(env, dict):
        env = {}
    # for default if no env HMAC method
    if not env or not env.get('auth_info', False):
        env['auth_info'] = {"auth_method": "CS HMAC"}

    # for public and private key in env
    if env.get('public_key', False) and env.get('private_key', False):
        env['auth_info'] = {"auth_method": "CS HMAC"}
        public_key = env.get('public_key')
        private_key = env.get('private_key')
    else:
        public_key = settings.APPLIANCE_PUBLIC_KEY
        private_key = settings.APPLIANCE_PRIVATE_KEY
    auth_info = env.get('auth_info')

    if not body and not method.lower() in bodyless_methods:
        body = public_key

    return body, auth_info, public_key, private_key


def make_request(url, method, body=None, *args, **kwargs):
    """
    This function facilitates using the crud hub api.

    It is for general purpose requests, but takes care of authentication
    automatically.

   :param str collection: An IRI that points to the location of the \
       crud hub collection (E.g. /api/3/events)
   :param str method: HTTP method
   :param dict body: An object to json encode and send to crud hub

   :return: the API response either as a json-like dict if possible or as bytes
   :rtype: dict/bytes
    """
    # ctrl c/v
    # get rid of the body on GET/HEAD requests

    # account for relative iri
    request_param_keys = ['params', 'data', 'headers', 'cookies', 'files', 'timeout', 'allow_redirects', 'proxies',
                          'stream', 'cert']
    request_params = {}

    if not bool(urlparse(url).netloc):
        url = settings.CRUD_HUB_URL + str(url)
    else:
        url = url

    bodyless_methods = ['head', 'get']
    if method.lower() in bodyless_methods:
        body = None

    if type(body) == str:
        try:
            body = ast.literal_eval(body)
        except Exception:
            pass

    logger.info('Starting request: %s , %s', method, url)

    for param_key in request_param_keys:
        param_value = kwargs.get(param_key)
        if param_value or isinstance(param_key, bool):
            request_params[param_key] = param_value

    body, auth_info, public_key, private_key = _get_request_params(method, body=body, *args, **kwargs)
    auth = get_requests_auth(auth_info, url, method,
                             public_key,
                             private_key,
                             json.dumps(body), *args, **kwargs)

    response = requests.request(method, url, auth=auth, json=body, verify=False, **request_params)
    if not kwargs.get('validate_response', True):
        return response
    return maybe_json_or_raise(response)


def make_file_upload_request(file_name, file_content, file_type, collection='files', *args, **kwargs):
    multipart_headers = kwargs.get('multipart_headers') or {'Expire': 0}
    request_headers = kwargs.get('request_headers') or {}
    extra_fields = kwargs.get('extra_multipart_fields') or {}
    url = settings.CRUD_HUB_URL + '/api/3/{0}'.format(collection)
    body, auth_info, public_key, private_key = _get_request_params("POST", *args, **kwargs)
    auth = get_requests_auth(auth_info, url, "POST",
                             public_key,
                             private_key,
                             body, *args, **kwargs)
    files = {'file': (file_name, file_content, file_type, multipart_headers)}
    files.update(extra_fields)
    response = requests.post(url, auth=auth, files=files, verify=False, headers=request_headers)
    return maybe_json_or_raise(response)


def maybe_json_or_raise(response):
    """
    Helper function for processing request responses

    Returns any json found in the response. Otherwise, it will extract the
    response as text, or, failing that, as bytes.

    :return: the response from the request
    :rtype: dict or str or bytes
    :raises: :class:`requests.HTTPError` if status code was 4xx or 5xx
    """
    if response.ok:
        try:
            logger.info('Processing request responses.')
            return response.json(strict=False)
        except Exception:
            return response.text or response.content
    else:
        msg = ''
        try:
            msg = response.json()
            logger.warn(msg)
        except Exception:
            pass
        if not msg:
            msg = response.text
            logger.warn(msg)

        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            # add any response content to the error message
            msg = '{} :: {} :: Url: {}'.format(str(e), msg, response.url)
            logger.error(msg)
            raise requests.exceptions.HTTPError(msg, response=response)
